==> coventry_0.5.3.orig.tar.gz: coventry-0.5.3/server/utils.cpp <==
/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "utils.hpp"
#include "events.hpp"

#include <strings.hpp>
#include <digest.hpp>

#include <cstring>
#include <fstream>

#include <sys/stat.h>

using namespace server;
using namespace crypto;

namespace {
inline constexpr char digits[17] = "0123456789abcdef";

auto u8_1b(const uint8_t *bp)
{
    return bp[0] == 0x09 || bp[0] == 0x0a || bp[0] == 0x0d || (bp[0] >= 0x20 && bp[0] <= 0x7e);
}

auto u8_2b(const uint8_t *bp)
{
     return bp[0] >= 0xc2 && bp[0] <= 0xdf && bp[1] >= 0x80 && bp[1] <= 0xbf;
}

auto u8_3b(const uint8_t *bp)
{
    if(bp[0] == 0xe0 && bp[1] >= 0xa0 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] >= 0xe1 && bp[0] <= 0xef && bp[0] != 0xed && bp[1] >=0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] == 0xed && bp[1] >= 0x80 && bp[1] <= 0x9f && bp[2] >= 0x80 && bp[2] < 0xbf)
        return true;

    return false;
}

auto u8_4b(const uint8_t *bp)
{
    if(bp[0] == 0xf0 && bp[1] >= 0x90 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] >= 0xf1 && bp[0] <= 0xf3 && bp[1] >= 0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] == 0xf4 && bp[1] >= 0x80 && bp[1] <= 0x8f && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    return false;
}

auto hex_decode(char ch)
{
    return static_cast<char>(isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10);
}

auto sha256_ha2(const event_t& event)
{
    std::string entity;
    if(event.qop() == "auth-int") {
        digest_t sha(EVP_sha256());
        osip_body_t *body = nullptr;
        if(event.request())
            osip_message_get_body(event.request(), 0, &body);
        if(body && body->body && body->length)
            sha.update(body->body, body->length);
        sha.finish();
        entity = util::to_hex(sha.data(), sha.size());
    }

    digest_t sha(EVP_sha256());
    sha.update(event.method(), util::count(event.method()));
    sha.update(":", 1);
    sha.update(event.uri().c_str(), event.uri().size());
    if(event.qop() == "auth-int") {
        sha.update(":", 1);
        sha.update(entity.data(), entity.size());
    }
    sha.finish();
    return util::to_hex(sha.data(), sha.size());
}

auto md5_ha2(const event_t& event)
{
    digest_t md5(EVP_md5());
    md5.update(event.method(), util::count(event.method()));
    md5.update(":", 1);
    md5.update(event.uri().c_str(), event.uri().size());
    md5.finish();
    return util::to_hex(md5.data(), md5.size());
}
} // end namespace anon

auto util::to_hex(const uint8_t *data, size_t size) -> std::string
{
    std::ostringstream out;
    while(data && size--) {
        out << digits[*data >> 4] << digits[*data & 0x0f];
        ++data;
    }
    return out.str();
}

auto util::from_hex(std::string_view from, uint8_t *to, size_t size) -> size_t
{
    auto count = 0U;
    while(count < size && from.size() > 1) {
        auto hi = hex_decode(from[0]);
        auto lo = hex_decode(from[1]);
        if(hi < 0 || lo < 0)
            break;
        *(to++) = static_cast<uint8_t>((hi << 4) | lo);
        from.remove_prefix(2);
        ++count;
    }
    return count;
}

auto util::get_value(std::string_view text) -> int
{
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto util::get_count(std::string_view count) -> uint32_t
{
    uint32_t value = 0;
    for(auto ch : count) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }
    return value;
}

auto util::get_expires(std::string_view cp) -> time_t
{
    if(cp.empty() || !isdigit(cp[0]))
        return 300;

    time_t value = 0;
    for(auto ch : cp) {
        if(ch == 's')
            return value;
        if(ch == 'm')
            return value * 60;
        if(ch == 'h')
            return value * 3600;
        if(ch < '0' || ch > '9')
            break;
        value *= 10;
        value += (ch - '0');
        if(value > 3600)    // 1 hour...
            return 3600;
    }
    if(value < 60)
        value = 60;
    return value;
}

auto util::get_period(const char *cp, unsigned period) -> unsigned
{
    if(!cp || !*cp)
        return period;

    unsigned value = 0;
    while(*cp) {
        if(*cp == 's' && !cp[1])
            return value;
        if(*cp == 'm' && !cp[1])
            return value * 60;
        if(*cp == 'h' && !cp[1])
            return value * 3600;
        if(*cp < '0' || *cp > '9')
            break;
        value *= 10;
        value += (*cp - '0');
        if(value > 31536000)    // 1 year...
            return period;
        ++cp;
    }
    return value;
}

auto util::get_port(const char *cp, uint16_t port) -> uint16_t
{
    if(!cp)
        return port;

    unsigned value = 0;
    while(*cp) {
        if(*cp < '0' || *cp > '9')
            break;
        value *= 10;
        value += (*cp - '0');
        if(value > 65535)
            return port;
        ++cp;
    }
    return static_cast<uint16_t>(value);
}

auto util::count(const char *cp, size_t max) -> size_t
{
    size_t count = 0;
    while(cp && *cp && count < max) {
        ++count;
        ++cp;
    }
    return count;
}

auto util::idverify(const char *cp, size_t max) -> size_t
{
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0;
    while(size < max && *cp) {
        if(!isalnum(*cp))
            return 0;
        ++size;
        ++cp;
    }

    if(size >= max)
        return 0;

    return size;
}

auto util::u8verify(std::string_view view) -> size_t
{
    size_t max = view.size(), size = 0UL;
    auto bp = reinterpret_cast<const uint8_t *>(view.data());

    while(size < max) {
        if(size < max && u8_1b(bp)) {
            ++size;
            ++bp;
            continue;
        }

        if(size + 1 < max && u8_2b(bp)) {
            size += 2;
            bp +=2;
            continue;
        }

        if(size + 2 < max && u8_3b(bp)) {
            size += 3;
            bp += 3;
            continue;
        }
        if(size + 3 < max && u8_4b(bp)) {
            size += 4;
            bp += 4;
            continue;
        }

        return 0;
    }

    return size;
}

auto util::u8verify(const char *cp, size_t max) -> size_t
{
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0, codepoints = 0;
    auto bp = reinterpret_cast<const uint8_t *>(cp);
    while(*bp && codepoints < max) {
        if(u8_1b(bp)) {
            ++bp;
            ++size;
            ++codepoints;
            continue;
        }

        if(u8_2b(bp)) {
            size += 2;
            bp +=2;
            ++codepoints;
            continue;
        }

        if(u8_3b(bp)) {
            size += 3;
            bp += 3;
            ++codepoints;
            continue;
        }
        if(u8_4b(bp)) {
            size += 4;
            bp += 4;
            ++codepoints;
            continue;
        }

        return 0;
    }

    if(codepoints > max || *bp)
        return 0;

    return size;
}

auto util::get_env(const char *id, size_t max) -> const char *
{
    const char *cp = ::getenv(id);  // FlawFinder: verified input
    if(!cp)
        return nullptr;

    if(!u8verify(cp, max))
        return nullptr;

    return cp;
}

auto util::sha256_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string
{
    digest_t sha(EVP_sha256());
    sha.update(userid.data(), userid.size());
    sha.update(":", 1);
    sha.update(realm.data(), realm.size());
    sha.update(":", 1);
    sha.update(secret.data(), secret.size());
    sha.finish();
    return util::to_hex(sha.data(), sha.size());
}

auto util::sha256_digest(std::string_view secret, const event_t& event) -> std::string
{
    auto ha2 = sha256_ha2(event);
    auto nonce = event.nonce();

    digest_t sha(EVP_sha256());
    sha.update(secret.data(), secret.size());
    sha.update(":", 1);
    sha.update(nonce.data(), nonce.size());
    sha.update(":", 1);

    if(event.qop() == "auth" || event.qop() == "auth-int") {
        sha.update(event.count().data(), event.count().size());
        sha.update(":", 1);
        sha.update(event.client().data(), event.client().size());
        sha.update(":", 1);
        sha.update(event.qop().data(), event.qop().size());
        sha.update(":", 1);
    }

    sha.update(ha2.data(), ha2.size());
    sha.finish();
    return util::to_hex(sha.data(), sha.size());
}

auto util::md5_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string
{
    digest_t md5(EVP_md5());
    md5.update(userid.data(), userid.size());
    md5.update(":", 1);
    md5.update(realm.data(), realm.size());
    md5.update(":", 1);
    md5.update(secret.data(), secret.size());
    md5.finish();
    return util::to_hex(md5.data(), md5.size());
}

auto util::md5_digest(std::string_view secret, const event_t& event) -> std::string
{
    auto ha2 = md5_ha2(event);
    auto nonce = event.nonce();

    digest_t md5(EVP_md5());
    md5.update(secret.data(), secret.size());
    md5.update(":", 1);
    md5.update(nonce.data(), nonce.size());
    md5.update(":", 1);
    md5.update(ha2.data(), ha2.size());
    md5.finish();
    return util::to_hex(md5.data(), md5.size());
}

auto util::is_bool(std::string value) -> bool
{
    value = lower_case(value);
    if(value == "t" || value == "f" || value == "n" || value == "y")
        return true;

    return value == "true" || value == "yes" || value == "1" || value == "false" || value == "no" || value == "0";
}

auto util::is_true(std::string_view value) -> bool
{
    if(value.empty())
        return false;

    return value[0] == '1' || tolower(value[0]) == 't' || tolower(value[0]) == 'y';
}

auto util::is_file(const std::string& path) -> bool
{
    struct stat ino{};
    if(::lstat(path.c_str(), &ino) < 0)
        return false;
    return S_ISREG(ino.st_mode);
}

auto util::set_string(char *cp, size_t max, std::string_view view) -> size_t
{
    assert(cp != nullptr);

    auto count = view.size();
    auto dp = view.data();
    if(count >= max)
        count = max - 1;

    max = count;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return count;
}

auto util::set_string(char *cp, size_t max, const char *from) -> size_t
{
    assert(cp != nullptr);
    assert(from != nullptr);
    assert(max > 0);

    auto count = 0UL;
    while(--max && *from) {
        *(cp++) = *(from++);
        ++count;
    }
    *cp = 0;
    return count;
}

auto util::get16(const uint8_t *ptr) -> uint16_t
{
    return ((uint16_t)(ptr[0]) << 8) | (uint16_t)(ptr[1]);
}

auto util::get32(const uint8_t *ptr) -> uint32_t
{
    return ((uint32_t)(ptr[0]) << 24) | ((uint32_t)(ptr[1]) << 16) |
        ((uint32_t)(ptr[2]) << 8) | (uint32_t)(ptr[3]);
}

auto util::get64(const uint8_t *ptr) -> uint64_t
{
    return ((uint64_t)(ptr[0]) << 56) | ((uint64_t)(ptr[1]) << 48) |
        ((uint64_t)(ptr[2]) << 40) | ((uint64_t)(ptr[3]) << 32) |
        ((uint64_t)(ptr[4]) << 24) | ((uint64_t)(ptr[5]) << 16) |
        ((uint64_t)(ptr[6]) << 8) | (uint64_t)(ptr[7]);
}

void util::set16(uint8_t *ptr, uint16_t value)
{
    ptr[0] = (value & 0xff00) >> 8;
    ptr[1] = value & 0xff;
}

void util::set32(uint8_t *ptr, uint32_t value)
{
    ptr[0] = (value & 0xff000000) >> 24;
    ptr[1] = (value & 0x00ff0000) >> 16;
    ptr[2] = (value & 0x0000ff00) >> 8;
    ptr[3] = value & 0xff;
}

void util::set64(uint8_t *ptr, uint64_t value)
{
    ptr[0] = (value & 0xff00000000000000) >> 56;
    ptr[1] = (value & 0x00ff000000000000) >> 48;
    ptr[2] = (value & 0x0000ff0000000000) >> 40;
    ptr[3] = (value & 0x000000ff00000000) >> 32;
    ptr[4] = (value & 0x00000000ff000000) >> 24;
    ptr[5] = (value & 0x0000000000ff0000) >> 16;
    ptr[6] = (value & 0x000000000000ff00) >> 8;
    ptr[7] = value & 0xff;
}

==> coventry_0.7.6.orig.tar.gz: coventry-0.7.6/server/utils.cpp <==
/*
 * Copyright (C) 2020-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "utils.hpp"
#include "events.hpp"

#include <strings.hpp>
#include <digest.hpp>

#include <cstring>
#include <fstream>

#include <sys/stat.h>
#include <pthread.h>

using namespace server;
using namespace crypto;

namespace {
auto u8_1b(const uint8_t *bp) {
    return bp[0] == 0x09 || bp[0] == 0x0a || bp[0] == 0x0d || (bp[0] >= 0x20 && bp[0] <= 0x7e);
}

auto u8_2b(const uint8_t *bp) {
     return bp[0] >= 0xc2 && bp[0] <= 0xdf && bp[1] >= 0x80 && bp[1] <= 0xbf;
}

auto u8_3b(const uint8_t *bp) {
    if(bp[0] == 0xe0 && bp[1] >= 0xa0 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] >= 0xe1 && bp[0] <= 0xef && bp[0] != 0xed && bp[1] >=0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] == 0xed && bp[1] >= 0x80 && bp[1] <= 0x9f && bp[2] >= 0x80 && bp[2] < 0xbf)
        return true;

    return false;
}

auto u8_4b(const uint8_t *bp) {
    if(bp[0] == 0xf0 && bp[1] >= 0x90 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] >= 0xf1 && bp[0] <= 0xf3 && bp[1] >= 0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] == 0xf4 && bp[1] >= 0x80 && bp[1] <= 0x8f && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    return false;
}

auto sha256_ha2(const event_t& event) {
    std::string entity;
    if(event.qop() == "auth-int") {
        digest_t sha(EVP_sha256());
        osip_body_t *body = nullptr;
        if(event.request())
            osip_message_get_body(event.request(), 0, &body);
        if(body && body->body && body->length)
            sha.update(body->body, body->length);
        sha.finish();
        entity = to_hex(sha.data(), sha.size());
    }

    digest_t sha(EVP_sha256());
    sha.update(event.method(), str_size(event.method()));
    sha.update(":", 1);
    sha.update(event.uri().c_str(), event.uri().size());
    if(event.qop() == "auth-int") {
        sha.update(":", 1);
        sha.update(entity.data(), entity.size());
    }
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto md5_ha2(const event_t& event) {
    digest_t md5(EVP_md5());
    md5.update(event.method(), str_size(event.method()));
    md5.update(":", 1);
    md5.update(event.uri().c_str(), event.uri().size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}
} // end namespace anon

auto util::get_value(std::string_view text) -> int {
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto util::get_count(std::string_view count) -> uint32_t {
    uint32_t value = 0;
    for(auto ch : count) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }
    return value;
}

auto util::get_expires(std::string_view cp) -> time_t {
    if(cp.empty() || !isdigit(cp[0]))
        return 300;

    time_t value = 0;
    for(auto ch : cp) {
        if(ch == 's')
            return value;
        if(ch == 'm')
            return value * 60;
        if(ch == 'h')
            return value * 3600;
        if(ch < '0' || ch > '9')
            break;
        value *= 10;
        value += (ch - '0');
        if(value > 3600)    // 1 hour...
            return 3600;
    }
    if(value < 60)
        value = 60;
    return value;
}

auto util::get_period(const char *cp, unsigned period) -> unsigned {
    if(!cp || !*cp)
        return period;

    unsigned value = 0;
    while(*cp) {
        if(*cp == 's' && !cp[1])
            return value;
        if(*cp == 'm' && !cp[1])
            return value * 60;
        if(*cp == 'h' && !cp[1])
            return value * 3600;
        if(*cp < '0' || *cp > '9')
            break;
        value *= 10;
        value += (*cp - '0');
        if(value > 31536000)    // 1 year...
            return period;
        ++cp;
    }
    return value;
}

auto util::get_port(const char *cp, uint16_t port) -> uint16_t {
    if(!cp)
        return port;

    unsigned value = 0;
    while(*cp) {
        if(*cp < '0' || *cp > '9')
            break;
        value *= 10;
        value += (*cp - '0');
        if(value > 65535)
            return port;
        ++cp;
    }
    return static_cast<uint16_t>(value);
}

auto util::idverify(const char *cp, size_t max) -> size_t {
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0;
    while(size < max && *cp) {
        if(!isalnum(*cp))
            return 0;
        ++size;
        ++cp;
    }

    if(size >= max)
        return 0;

    return size;
}

auto util::u8verify(std::string_view view) -> size_t {
    const auto max = view.size();
    auto size = size_t(0UL);
    auto bp = reinterpret_cast<const uint8_t *>(view.data());

    while(size < max) {
        if(u8_1b(bp)) {
            ++size;
            ++bp;
            continue;
        }

        if(size + 1 < max && u8_2b(bp)) {
            size += 2;
            bp +=2;
            continue;
        }

        if(size + 2 < max && u8_3b(bp)) {
            size += 3;
            bp += 3;
            continue;
        }
        if(size + 3 < max && u8_4b(bp)) {
            size += 4;
            bp += 4;
            continue;
        }

        return 0;
    }

    return size;
}

auto util::u8verify(const char *cp, size_t max) -> size_t {
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0, codepoints = 0;
    auto bp = reinterpret_cast<const uint8_t *>(cp);
    while(*bp && codepoints < max) {
        if(u8_1b(bp)) {
            ++bp;
            ++size;
            ++codepoints;
            continue;
        }

        if(u8_2b(bp)) {
            size += 2;
            bp +=2;
            ++codepoints;
            continue;
        }

        if(u8_3b(bp)) {
            size += 3;
            bp += 3;
            ++codepoints;
            continue;
        }
        if(u8_4b(bp)) {
            size += 4;
            bp += 4;
            ++codepoints;
            continue;
        }

        return 0;
    }

    if(codepoints > max || *bp)
        return 0;

    return size;
}

auto util::get_env(const char *id, size_t max) -> const char * {
    const char *cp = ::getenv(id);  // FlawFinder: verified input
    if(!cp)
        return nullptr;

    if(!u8verify(cp, max))
        return nullptr;

    return cp;
}

auto util::sha256_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string {
    digest_t sha(EVP_sha256());
    sha.update(userid.data(), userid.size());
    sha.update(":", 1);
    sha.update(realm.data(), realm.size());
    sha.update(":", 1);
    sha.update(secret.data(), secret.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto util::sha256_digest(std::string_view secret, const event_t& event) -> std::string {
    auto ha2 = sha256_ha2(event);
    auto nonce = event.nonce();

    digest_t sha(EVP_sha256());
    sha.update(secret.data(), secret.size());
    sha.update(":", 1);
    sha.update(nonce.data(), nonce.size());
    sha.update(":", 1);

    if(event.qop() == "auth" || event.qop() == "auth-int") {
        sha.update(event.count().data(), event.count().size());
        sha.update(":", 1);
        sha.update(event.client().data(), event.client().size());
        sha.update(":", 1);
        sha.update(event.qop().data(), event.qop().size());
        sha.update(":", 1);
    }

    sha.update(ha2.data(), ha2.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto util::md5_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string {
    digest_t md5(EVP_md5());
    md5.update(userid.data(), userid.size());
    md5.update(":", 1);
    md5.update(realm.data(), realm.size());
    md5.update(":", 1);
    md5.update(secret.data(), secret.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

auto util::md5_digest(std::string_view secret, const event_t& event) -> std::string {
    auto ha2 = md5_ha2(event);
    auto nonce = event.nonce();

    digest_t md5(EVP_md5());
    md5.update(secret.data(), secret.size());
    md5.update(":", 1);
    md5.update(nonce.data(), nonce.size());
    md5.update(":", 1);
    md5.update(ha2.data(), ha2.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

auto util::is_bool(std::string value) -> bool {
    value = lower_case(value);
    if(value == "t" || value == "f" || value == "n" || value == "y")
        return true;

    return value == "true" || value == "yes" || value == "1" || value == "false" || value == "no" || value == "0";
}

auto util::is_true(std::string_view value) -> bool {
    if(value.empty())
        return false;

    return value[0] == '1' || tolower(value[0]) == 't' || tolower(value[0]) == 'y';
}

auto util::set_priority(int priority) -> bool {
    auto tid = pthread_self();
    struct sched_param sp{};
    int policy{};

    std::memset(&sp, 0, sizeof(sp));

    policy = SCHED_OTHER;
#ifdef  SCHED_FIFO
    if(priority > 0) {
        policy = SCHED_FIFO;
        priority = sched_get_priority_min(policy) + priority - 1;
        if(priority > sched_get_priority_max(policy))
            priority = sched_get_priority_max(policy);
    }
#endif
#ifdef  SCHED_BATCH
    if(priority < 0) {
        policy = SCHED_BATCH;
        priority = 0;
    }
#endif
    if(policy == SCHED_OTHER)
        priority = 0;

    sp.sched_priority = priority;
    return pthread_setschedparam(tid, policy, &sp) == 0;
}
